(* struct CLOS *)

open Pie
open Exp
open Surr
   
type surr = env * cxt
type clos = surr * exp

(* Surroundings *)
	      
let empty = (env_empty, cxt_empty)
let push  (x, t, (theta, gamma)) = (theta, cxt_push  (x, t, (theta, gamma)))
let pushTmVar (x, (gR, tau), (thL, gL)) = (thL, cxt_pushTmVar (x, (gR, tau), (thL, gL)))
let pushTyVar (a, kappa, (thL, gL)) = (thL, cxt_pushTyVar (a, kappa, (thL, gL)))
let push' (x, (theta, gamma))    = (theta, cxt_push' (x, (theta, gamma)))
let pushLit' (x, (theta, gamma)) = (theta, cxt_pushLit' (x, gamma))
				     
exception UnknownIdent of string
  
let index ((theta, gamma), x) =
  match cxt_index (gamma, x) with
      None -> (match env_index (theta, x) with
		   None -> raise (UnknownIdent x)
		 | Some i -> i)
    | Some i -> i
	

(* Closures *)

let close (t) = (empty, t)
		  
(* precedence:
   10: app right
   9: app left
   5: -> left
   4: -> right
   0: forall, lam
*)
	
let toString' (n0, ((theta, gamma), t)) =
 
  (** [tos: bool * int * cxt * exp -> string] 

    [tos (b, n, g, t') = s]

    [b = true] means we are in rightmost position.  This enables us to
    drop some parenthesis and would even enable extensive use of the
    "." notation.  We only will use it for omitting parenthesis around
    a rightmost "forall".
  *)
  let rec tos (b, n, g, t') =
    let par () = 
      let s = tos (true, 0, g, t') in
	if !Output.extensiveDot && b then ". "^s
	else "("^ s ^")"
    in
    match t' with

	Var i -> (if i >= 0 
		  then let (x, t) = cxt_lookup (g, i) in x
		  else match (env_lookup (theta, i)) with
		      Def (x, tau, t, strict) -> x
		    | Decl (x, tau) -> x
		    | Cell c -> 
			match !c with
			    Name (x, g', kappa) -> x
			  | Value (tau, g', kappa) -> tos (b, n, g', tau))

      | C(Type, []) -> Keyword.star

      | C(Kind, []) -> Keyword.kind

      | C(App, [[],s;[],t]) -> if n >= 10 then par ()
	else (tos (false, 9, g, s))^" "^(tos (b, 10, g, t))

      | C(Arr, [[],s;[],t]) -> if n >= 5 then par ()
	else (tos (false, 5, g, s))^" "^Keyword.arr^" "^(tos (b, 4, g, t))

      | C(Lam, [[x],t]) -> if n > 0 && (not b) then par ()
	else let x' = nextInst ((theta, g), x)
	in Keyword.lam^ x' ^ (lamAll (false, cxt_pushLit' (x', g), t))

      | C(Forall, [[],s; [x],t]) -> if n > 0 && (not b) then par ()
	else let  x' = nextInst ((theta, g), x)
	in Keyword.all^" "^x'^":"^(tos (false, 1, g, s))^(lamAll (true, cxt_pushLit' (x', g), t))

      | _ -> "MALFORMED EXPRESSION"

  (** [lamAll]: special printing of a sequence of lambdas/foralls:
    only print "." after last lambda/forall *)

  and lamAll (needSpace, g, t') =
    match t' with
	C(Lam, [[x],t]) ->
	  let x' = nextInst ((theta, g), x)
	  in (if needSpace then " " else "") ^ 
	     Keyword.lam ^ x' ^ 
	     (lamAll (false, cxt_pushLit' (x', g), t))
      | C(Forall, [[],s; [x],t]) -> 
	  let  x' = nextInst ((theta, g), x)
	  in " " ^ Keyword.all ^ " " ^x' ^ ":" ^ (tos (false, 1, g, s)) ^ 
	     (lamAll (true, cxt_pushLit' (x', g), t))
      | _ -> ". " ^ (tos (true, 0, g, t'))
	  
  in (* begin toString' *)
    if n0=0 then tos (true, 0, gamma, t)
    else tos (false, n0, gamma, t)

let toString ((theta, gamma), t) = toString' (0, ((theta, gamma), t))
