(** A parser for command line arguments *)

open Output
open Genlex

exception Error of string

type option =
    Num of int           (** --opt=<num>                 *)
  | List of string list  (** --opt=<string>,...,<string> *)
  | Nothing              (** --opt                       *)
    
type arg =
    ShortOption of string         (** -o             *)
  | LongOption of string * option (** --opt=value    *)
  | File of string     

let unrecognizedOption  (a) = raise (Error ("unrecognized option \""^a^"\""))
let wrongSyntaxInOption (a) = raise (Error ("wrong syntax in option \""^a^"\""))

let lexer = make_lexer ["-"; "--"; "="; ","]

(** top function for parsing *)
let parseArg (a) =
  let s = lexer (Stream.of_string (a)) in

  (** [cont (success, failure)]: continuation of parsing.
  Reads next token, if token is present, then call [success] else call 
  [failure] *)
  let cont (success, failure) =
    try
      success (Stream.next (s))
    with
	Stream.Failure -> failure ()

  and unrecognizedOpt _ = unrecognizedOption (a)
  and wrongSyntax _ = wrongSyntaxInOption (a)

  in let parseShortOpt (t) =
      match t with
	  Ident (x) -> cont (wrongSyntax, fun () -> ShortOption (x))
	| _ -> unrecognizedOpt ()

     and parseLongOpt (t) =

      let parseEquals x (t) =

	let parseValue (t) =

	  let returnList vs _ = LongOption (x, List (List.rev vs)) in
	   
	  let rec parseComma vs (t) =
	    match t with
		Kwd "," -> cont (parseNextId vs, wrongSyntax)
	      | _ -> wrongSyntax ()

	  and parseNextId vs (t) =
	    match t with
		Ident v -> cont (parseComma (v::vs), returnList (v::vs))
	      | _ -> wrongSyntax ()
		
(* does not work  
	  and parseInt (t) =
	    match t with
		Int n -> cont (wrongSyntax, fun () -> LongOption (x, Num (-n)))
	      | _ -> wrongSyntax ()
*)

	  in (* begin parseValue *)
	    match t with
		Int n -> cont (wrongSyntax, fun () -> LongOption (x, Num n))
(*
	      | Kwd "-" -> cont (parseInt, wrongSyntax)
*)
	      | Ident v0 -> cont (parseComma [v0], returnList [v0])
	      | _ -> wrongSyntax ()

	in (* begin parseEquals *)
	  match t with
	      Kwd "=" -> cont (parseValue, fun () -> LongOption (x, List []))
	    | _ -> wrongSyntax ()

      in (* begin parseLongOpt *)
	match t with 
	    Ident (x) -> cont (parseEquals x, 
			       fun () -> LongOption (x, Nothing))
	  | _ -> unrecognizedOpt ()

  (** initial parser, detects whether option or not *)
  in let parse0 (t) =
      match t with
	  Kwd "-" -> cont (parseShortOpt, unrecognizedOpt) 
	| Kwd "--" -> cont (parseLongOpt, unrecognizedOpt)
	| _ -> File (a)
	    
  in (* begin parseArg *)
    cont (parse0, fun () -> raise (Error ("empty argument \""^a^"\"")))
