(** Kind-driven beta-etq-Equality for constructors. *)

open Output
open Pie
open Exp
open Surr
open Norm

exception Fail of string  (** pseudo exception if equality test fails *)
(* exception Error of string (** real exception *) *)

(** Equality of constructors
    [eq ((th, g), f01, f02, kappa0) = kappa0] 

Currently the implementation is brute-force, expanding all definitions using [hnf].  A more refined implementation would only expand by need.
*)

type kdOrStr =
    StrEq            (** Structural equality *)
  | KDrEq of Exp.exp (** Kind-driven equality with parameter [kappa] *)

let rec kDrOrStrEq ((th, g), f01, f02, kDrOrStr) =
(*
  log ("enter_eq", fun () -> 
	 "? "^Clos.toString ((th, g), f01)^" == "^Clos.toString ((th, g), f02)^"\n");
*)

  (** STRuctural EQuality 
    [streq (h1, h2) = kappa] *)
  let rec streq (h1, h2) =
    
    log ("enter_strec", fun () ->
	 "? strec: "^Clos.toString ((th, g), h1)^" == "^Clos.toString ((th, g), h2)^"\n");
    
    let notEqual () =
      raise(Fail("Checking equality of the following constructors failed.\n"^
		 "LHS: "^Clos.toString ((th, g), f01)^"\n"^
		 "RHS: "^Clos.toString ((th, g), f02)^"\n"^
		 "The following subterms are inequal: "^
		 Clos.toString ((th, g), h1)^"\n  and "^
		 Clos.toString ((th, g), h2)^"\n"))
	
    in (* begin strec *)
      match (h1, h2) with
	  
	  (* case: two variables *)
	  (Var i, Var j) ->
	    if (i = j) then 
	      if i >= 0 then
		let (a, kappa) = cxt_lookupTyVar (g, i)
		in kappa
	      else match env_lookup (th, i) with
		  Decl (a, kappa) -> kappa
		| Def (a, kappa, tau, strict) -> kappa
		| Cell (c) -> match !c with
		      Name (a, gamma, kappa) -> kappa
		    | Value (tau, gamma, kappa) -> kappa
(*
		    (* other cases impossible because of hnf *)
		| _ -> raise (Error "head is unexpanded definition or uninst. EVar (despite hnf)")
*)
	    else notEqual ()

	  (* case: two stars *)
	| (C(Type, []), C(Type, [])) -> 
	    (* for using streq with kinds *)
	    mkKind 

	  (* case: two arrows *)
	| (C(Arr, [[],sigma1; [],tau1]), C(Arr, [[],sigma2; [],tau2])) ->
	    (* recursively continue componentwise *)
	    streq (hnf ((th, g), sigma1), hnf ((th, g), sigma2));
	    streq (hnf ((th, g), tau1), hnf ((th, g), tau2)) 
	    
	  (* case: two universal quantifications *)
	| (C(Forall, [[],k1; [a1],tau1]), C(Forall, [[],k2; [a2],tau2])) ->
	    (* first compare kinds *)
	    ignore (streq (k1, k2));
	    (* then treat them as lambdas *)
	    kdreq (g, mkLam (a1, tau1), mkLam (a2, tau2), mkArr (k1, mkType))

          (* case: two neutral terms (applications) *)
	| (C(App, [[],f1; [],a1]), C(App, [[],f2; [],a2])) ->
	    (* compare heads and obtain arrow kind k->k' *)
	    (match streq (f1, f2) with
		C(Arr, [[],k; [],k']) ->
		  (* compare arguments at kind k *)
		  ignore (kdreq (g, a1, a2, k)); 
		  (* return kind k' *)
		  k'

	      | _ -> raise (Error "type of head in neutral application is not Arr"))
	    
	| _ -> notEqual ()
	    
	    

  (** Kind-DRiven-EQuality 
      [kdreq (g, f1, f2, kappa) = ()] *)
  and kdreq (g, f1, f2, kappa) =

    let streq (h1, h2) = kDrOrStrEq ((th, g), h1, h2, StrEq)

    in (* begin kdreq *)
      match kappa with
	
      (* case: kappa is function kind k1->k2 *)
	C(Arr, [[],k1;[],k2]) ->
	  (* assume new constructor variable A *)
	  let g' = cxt_pushTyVar ("A", k1, (th, g)) in
	    (* remember to lift f1 and f2 *)
	  let f1' = Pie.lift1 f1 in
	  let f2' = Pie.lift1 f2 in
	    (* recursively check equality of (f1' A) and (f2' A) *)
	    kdreq (g', mkApp (f1', Pie.Var 0), mkApp (f2', Pie.Var 0), k2) 

      (* case: kappa is basic kind * *)
      | C(Type, []) -> 
	  (* continue with structural equality of head normal expressions *)
	  streq (hnf ((th, g), f1), hnf ((th, g), f2))

      (* case for equality of kinds *)
      | C(Kind, []) -> 
	  (* continue with structural equality of head normal expressions *)
	  streq (hnf ((th, g), f1), hnf ((th, g), f2))

  in (* begin eq *)
    match kDrOrStr with
	StrEq -> streq (f01, f02)
      | KDrEq (kappa) -> kdreq (g, f01, f02, kappa); kappa

   
let eq ((th, g), f1, f2, kappa0) =

  log ("enter_eq", fun () -> 
	 "? "^Clos.toString ((th, g), f1)^" == "^Clos.toString ((th, g), f2)^"\n");
  ignore (kDrOrStrEq ((th, g), f1, f2, KDrEq (kappa0)))

let streq ((th, g), f1, f2) =
  ignore (kDrOrStrEq ((th, g), f1, f2, StrEq))
