(* struct Exp *)

open Pie

exception Error of string (** general purpose exception *)

(* Expressions *)

type expC = Lam | App | Arr | Forall | Type | Kind 
	  | TmAbs | TmApp | TyAbs | TyApp (* Church Style Terms *)
type exp  = expC pieTm
	      
let rec wfKind e =
  match e with
      C(Type, []) -> true
    | C(Arr, [[],f; [],g]) -> wfKind f && wfKind g
    | _ -> false 

let mkApp (s, t)    = C(App, [[],s; [],t]) ;;
let mkLam (x, t)    = C(Lam, [[x],t]) ;;

let mkArr (s, t)    = C(Arr, [[],s; [],t]) ;;
let mkForall (a, s, t) = C(Forall, [[],s; [a],t]) ;; 

let mkType          = C(Type, [])
let mkKind          = C(Kind, [])


let mkTmAbs (x, tau, t)   = C (TmAbs, [[],tau; [x],t])
let mkTyAbs (a, kappa, t) = C (TyAbs, [[],kappa; [a],t])

let mkTmApp (s, t)        = C (TmApp, [[],s; [],t])
let mkTyApp (t, tau)      = C (TyApp, [[],t; [],tau])

let cToString c = 
  match c with
      Lam -> "Lam"
    | App -> "App"
    | Arr -> "Arr"
    | Forall -> "Forall"
    | Type -> "Type"
    | Kind -> "Kind"
    | TmAbs -> "TmAbs"
    | TmApp -> "TmApp"
    | TyAbs -> "TyAbs"
    | TyApp -> "TyApp"

let toString t = Pie.toString cToString t


(** Lifting for Church-style terms.

Because in Church-style terms, term and type variables are distinguished, the [Pie.lift] function cannot be used for Church-terms.  [Pie.lift] count all binders as binders for all variables.  This would be ok if we used a mixed term/type context (actually the better choice), but since we have separate contexts we also need separate lifting. *)

let rec tmVarLift (t0, l, n) = 

  match t0 with

      (* term variable *)
      Var k -> if k < l then Var k else Var (k+n)

      (* term var abstraction (count) *)
    | C(TmAbs, [[],tau; [x],t]) ->
	mkTmAbs (x, tau, tmVarLift (t, l+1, n))

      (* term application *)
    | C(TmApp, [[],r; [],s]) ->
	mkTmApp (tmVarLift (r, l, n), tmVarLift (s, l, n))

      (* type var abstraction (do not count!) *)
    | C(TyAbs, [[],kappa; [a],t]) ->
	mkTyAbs (a, kappa, tmVarLift (t, l, n))

      (* type application: right component is tm var free *)
    | C(TyApp, [[],r; [],tau]) ->
	mkTyApp (tmVarLift (r, l, n), tau)

    | _ -> raise (Error ("ill-formed Church-term"^toString(t0)))

let tmVarLift1 (t) = tmVarLift (t, 0, 1)


let rec tyVarLift (t0, l, n) =

  match t0 with

      (* term variable: do not lift *)
      Var k -> Var k

      (* term var astraction: do not count, but lift type in annotation *)
    | C(TmAbs, [[],tau; [x],t]) ->
	mkTmAbs (x, liftTm (tau, l, n), tyVarLift (t, l, n))

      (* term application *)    
    | C(TmApp, [[],r; [],s]) ->
	mkTmApp (tyVarLift (r, l, n), tyVarLift (s, l, n))

      (* type var abstraction (count) *)
    | C(TyAbs, [[],kappa; [a],t]) ->
	mkTyAbs (a, kappa, tmVarLift (t, l + 1, n))

      (* type application: std-lift right component *)
    | C(TyApp, [[],r; [],tau]) ->
	mkTyApp (tmVarLift (r, l, n), liftTm (tau, l, n))

    | _ -> raise (Error ("ill-formed Church-term"^toString(t0)))

let tyVarLift1 (t) = tyVarLift (t, 0, 1)
