(** Main program. *)

open Commandline
open Output

let myself = "fomega"
let version = "%% Fomega Version 0.10\n%% A type checker for (some) annotation-free terms of System F^omega.\n"

(** Options *)

let debugTrace = ref false

(** Benchmarking *)

let bench = ref true        (** print benchmark ? *)
let startTime = Sys.time () (** get system time at start *)

let printBench () =
  if !bench then 
(*
    prerr_string ("Elapsed time: "^
		  string_of_float (Sys.time () -. startTime)^" sec\n")
*)
    Printf.eprintf "Elapsed time : %.2f sec\n" (Sys.time () -. startTime)
  else ()

let doStream s = Parser.parseFile (Parser.lexer (Stream.of_channel s))

let rec run (files) =
  match files with
      [] -> ()
    | (f::fs) ->
	let logFile = (try
			 Filename.chop_extension (f)
		       with Invalid_argument _ -> f)^".log" in
	  startLog (logFile);
	  log ("file_msg", fun () -> "%% Opening file: "^f^"\n");
	  doStream (open_in (f));
	  log ("file_msg", fun () -> "\n%% Closing file: "^f^"\n");  
	  run (fs)

let usage () =
  print_string (
    "usage: "^myself^" [options] files\n\n"^
    "Options:\n\n"^
    "    --help                : display this message and exit\n\n"^
    "    --stdin               : read input from stdin instead of files\n\n"^
    "    --church=on|off       : print Church terms (default: on)\n\n"^
    "    --extensivedot=on|off : make extensive use of dot-notation (default: off)\n\n"^
    "    --verbosity=<num>     : set verbosity level (default: 2)\n\n"^
    "        level 0           : feedback on term and type declarations\n"^
    "        level 1           : summarizing steps of type checking\n"^
    "        level 2           : main steps of type checking and unification\n"^   
    "        level 3           : church terms in type checking and eta red. steps\n"^
    "        level 4           : messages on entering major functions\n"^
    "        level 5           : messages on entering minor functions\n"^
    "        level 6           : all, including debug messages\n\n"^
    "        messages with level >= 1 are only printed if something goes wrong\n\n"^
    "    --debug               : send messages to stderr instead of stdout\n\n"^
    "    --log=<num>           : while processing file.syf, log into file.log\n"^
    "                            with verbosity level <num>; messages are logged\n"^
    "                            even if nothing goes wrong\n\n")

let exitUsage (msg) =
  print_string (version^"\n");
  if msg = "" then () else print_string (msg^"\n\n");
  usage ();
  exit 1

let readFromStdin = ref false

let parseArgs (args) =

  let doArg files arg =
    let a = parseArg arg in
      match a with
	  File f -> f::files
	| _ -> 
	    (match a with
		 LongOption ("help", v) ->
		   (match v with
			Nothing -> exitUsage ("")
		      | _ -> wrongSyntaxInOption (arg))
	       | LongOption ("stdin", v) ->
		   (match v with
			Nothing -> readFromStdin := true
		      | _ -> wrongSyntaxInOption (arg))
	       | LongOption ("church", v) ->
		   (match v with
			List ["on"] -> Output.churchOn ()
		      | List ["off"] -> Output.churchOff ()
		      | _ -> wrongSyntaxInOption (arg))
	       | LongOption ("extensivedot", v) ->
		   (match v with
			List ["on"] -> Output.extensiveDot := true
		      | List ["off"] -> Output.extensiveDot := false
		      | _ -> wrongSyntaxInOption (arg))
	       | LongOption ("verbosity", v) ->
		   (match v with
			Num l -> Output.traceLevel := l
		      | _ -> wrongSyntaxInOption (arg))
	       | LongOption ("log", v) ->
		   (match v with
			Num l -> Output.logLevel := l
		      | _ -> wrongSyntaxInOption (arg))
	       | LongOption ("debug", v) ->
		   (match v with
			Nothing -> Output.debugOn ()
		      | _ -> wrongSyntaxInOption (arg))
	       | LongOption ("debugtrace", v) ->
		   (match v with
			List ["on"] -> debugTrace := true
		      | List ["off"] -> debugTrace := false
		      | _ -> wrongSyntaxInOption (arg))
	       | _ -> unrecognizedOption (arg)
	    ); files
	  
  in (* begin parseArgs *)
    List.rev (List.fold_left doArg [] args)
   
let abort () =
  printBench ();
  exit 1

let print_ex (msg) = log ("exception", fun () -> msg)

let printTrace () =
  try 
    Output.printTrace() 
  with ex' -> print_ex ("Exception during trace output.\n");
     if !debugTrace then raise ex' 
     else print_ex ("Continuing... (Exception can be seen with option '--debugtrace=on')\n")

let main () =
  let (arg0 :: args) = Array.to_list (Sys.argv) in
    if args = [] then exitUsage ("No arguments given.")
    else
  let files =
    try 
      parseArgs (args)
    with
	Commandline.Error msg -> exitUsage (msg)
  in
    if files = [] then
      if !readFromStdin then ()
      else exitUsage ("No files given.")
    else 
      if !readFromStdin then 
	exitUsage("Cannot use option '--stdin' with input files.")
      else ();
    Output.log ("version", fun () -> version);
    try
      if !readFromStdin then
	doStream (stdin)
        (* logging is to stderr by default *)
      else run (files);
      printBench ()
    with
	(Lexer.Fail (msg)) as ex ->
	  print_ex ("Parsing failed: "^msg^".\n");
	  abort ()
      | (Parser.Fail (msg)) as ex ->
	  print_ex ("Parsing failed: "^msg^".\n");
	  abort ()
      | (Uni.Fail (msg)) as ex -> 
	  printTrace ();
	  print_ex ("Unification failed for the following reason:\n"^msg);
	  abort ()
      | (Check.Fail (msg)) as ex -> 
	  printTrace ();
	  print_ex ("Type checking failed for the following reason:\n"^msg);
	  abort ()
      | (Eq.Fail (msg)) as ex ->
	  printTrace ();
	  print_ex (msg);
	  abort ()
      | (Output.Error (msg)) as ex  ->
	  printTrace ();
	  print_ex ("Internal Error:\n"^msg);
	  abort ()
      | ex -> 
	  printBench ();
	  flush (stdout); (* does this help anything ? *)
	  raise ex

;;
main ();; 



