(** Name handling. *)

type absName = string * int
(** An abstract name is a pair (stem: string, inst:int) of a string [stem], 
the base name, and a natural number [inst], the current instance *)

exception Error of string

let primeC = '\''
let primeS = "'"

(** [parse : string -> absName] *)
let parse (x) = 
  try
    (* ppos is the rightmost occurence of a prime in x *)
    let ppos = String.rindex x primeC in
    let error () = raise (Error "invalid variable name: begins with a ' (prime)") in
      if ppos <= 0 then error () else
	(* len is the length of the string after the last prime *)
	let len =  String.length x - (ppos+1) in
	  if len = 0 then
	    (* check for second prime right before the last prime *)
	    if String.get x (ppos-1) = primeC then
	      if ppos <= 1 then error () else
		(* check for third prime right before the second prime *)
		if String.get x (ppos-2) = primeC then
		  (* x is 3rd instance *)
		  if ppos <= 2 then error () else 
		    (String.sub x 0 (ppos-2), 3)
		else (* x is a 2nd instance *)
		  (String.sub x 0 (ppos-1), 2)
	    else (* x is a 1st instance *)
	      (String.sub x 0 ppos, 1)
	  else
	    (* (non-empty) tail is the suffix of x after the last prime *)
	    let tail = String.sub x (ppos+1) len in
	      try
		let i = int_of_string tail in
		  (* tail number must be >= 4 *)
		  if i < 4 then (x, 0)
		  else (String.sub x 0 ppos, i)
	      with
		  (* tail is no number: the whole of x is stem *)
		  Failure _ -> (x, 0)
  with
      (* if x contains no prime, then the whole of x is stem *) 
      Not_found -> (x, 0)
    

(** [unparse: absName -> string] *)
let unparse (stem, inst) = 
  if inst = 0 then stem
  else if inst <= 3 then stem ^ (String.make inst primeC)
  else if inst < 0 then raise (Error "negative instance in variable detected")
  else stem ^ primeS ^ string_of_int (inst)

(** [isInst : string * string -> bool] 
    [isInst (x, x')] checks whether [x'] is an instance of [x] *)
let isInst (x, x') =
  let (s, i) = parse (x)
  and (s', i') = parse (x')
  in s = s' && i' >= i

(** [nextInst : string -> string]
    [nextInst (x) = x'] where [x'] is the next instance wrt. [x] *)
let nextInst (x) = 
  let (s, i) = parse (x)
  in unparse (s, i+1)


(* Simple minded name instances:

let trailPrimes (s, start) =
  let rec iter (i) =
    i >= String.length(s) 
    || (String.get s i = primeC && iter (i+1))
  in
    iter (start)

(** [inInst (x, x') = true] if [x'] is [x] plus a number of primes *)
let isInst (x, x') =
  let lx = String.length x in
  let lx' = String.length x' in
    lx <= lx' && x = String.sub x' 0 lx && trailPrimes (x', lx)

(** instances of x: x' x'' x''' x'''' ... *)
let nextInst (x) = x ^ primeS

*)
