(** Functions for output, logging and debugging. *)

exception Error of string

(** Width of identifier field in output. *)
let nameWidth = 12

let nameField = String.make nameWidth ' '

let formatWidth w x = 
  let n = w - String.length (x) in 
    if n > 0 then x^(String.make n ' ')
    else x

let formatName x = formatWidth nameWidth x

let indent = 2

let extensiveDot = ref false

(** An assoc list of log_ids and their levels *)
let ids = [
  "version", -1;
  "exception", -1;
  "file_msg", -1;
  "decl", -1; (* (term or type) (decl or def) *)
  "church", 0;
  "adding_term_decl", 1;
  "adding_term_def", 1;
  "adding_type_decl", 1;
  "adding_type_def", 1;
  "strict", 1;
  "checking_wf_type", 1;
  "checking", 1;
  "normal_form", 1;
  "church_term", 1;
  "church_term_deBruijn", 6;
  "inferred_type", 1;
  "eta_church_term", 1;
  "eta_inferred_type", 1;
  "chk_chk", 1;
  "ass_chk", 1;
  "new_chk", 2;
  "set_chk", 2;
  "uni_chk", 2;
  "eq_chk", 2;
  "ass_uni", 2;
  "new_uni", 2;
  "set_uni", 2;
  "church_chk", 3;
  "church_uni", 4;
  "red_eta", 3;
  "enter_chk", 4;
  "enter_leq", 4;
  "enter_strict", 4;
  "enter_uniUp", 4;
  "enter_flex", 4;
  "debug_leq", 6;
  "exit_findHead", 4;
  "exit_infer", 5;
  "exit_check", 5;
  "enter_check", 5;
  "enter_eq", 5;
  "enter_strec", 5;
  "enter_eta", 5;
  "enter_eta_deBruijn", 6;
  "enter_norm", 5;
  "debug_eta", 6;
  "debug_norm", 6;
  "enter_churchPromote", 6;
  "enter_promote", 6
]

(* unused:  
let enabled = ref []
*)

let logLevel = ref (-2)
let logFile  = ref (stderr)

let outLevel = ref 0

(*
let debugLevel = ref (-2)
let debugOn (level) =
  debugLevel := level;
  outLevel := -2;
  traceLevel := -2
*)

let my_print_string = ref (print_string)

let debugOn () = my_print_string := 
		 fun (msg) -> prerr_string (msg); flush (stderr)

let churchOn  () = outLevel := 0
let churchOff () = outLevel := -1

let traceLevel = ref 2
let trace = ref ([] : string list)
(* lazy does not work due to references in env
let trace = ref ([] : (string Lazy.t) list)
*)
let startLog (filename) =
  if !logLevel > -2 then logFile := open_out (filename) else ()

let beginTrace () = trace := []

let log (id, msg) = 
  let msg' = Lazy.lazy_from_fun msg in
  try
    let lev =  List.assoc id ids in

      (* Logging into file *)
      if (* unused: List.mem id !enabled || *)
	lev <= !logLevel && not (id = "version") then 
	output_string !logFile (Lazy.force msg')
      else ();

      (* Ordinary output and trace *)
      if lev <= !outLevel then !my_print_string (Lazy.force msg')
      else if lev <= !traceLevel (* && !logLevel < !traceLevel *) then
	trace := Lazy.force msg' :: !trace 
(* 	trace := Lazy.lazy_from_fun msg :: !trace *)
     else ()

  with
      Not_found -> raise (Error ("log: invalid log classifier \""^id^"\""))

let printTrace () =
  List.fold_right (fun msg -> fun () -> !my_print_string (msg)) !trace ()
(*  List.fold_right (fun msg -> fun () -> print_string (Lazy.force msg)) !trace () *)
