(* struct Parser *)
(** Parse declarations. *)

open Pie
open Exp
open Clos
open Genlex 

(* exception Error of string *)
exception Fail of string

let lexer = Lexer.lexer

let parseFile (s) =

  (** [cont] reads next token and calls parser [success] 
      resp. [failure] if there are no more tokens. *)
  let cont (success: token -> 'a) (failure: unit -> 'a) =
    try
      success (Stream.next (s))
    with
	Stream.Failure -> failure () in

  (* [cont': (token -> 'a) -> 'a] *)
  let cont' (p: token -> 'a) =
    try
      p (Stream.next (s))
    with
	Stream.Failure -> raise (Fail "unexpected end of file") in

(*
  let contRParen' (p) =
    let error () = raise (Fail ("closing parenthesis ')' expected")) in
      try
	match (Stream.next ()) with
	    Kwd ")" -> cont' (p)
	  | _ -> error ()
	with
	    Stream.Failure -> error ()
*)	      

  (** [makeArr [t0; t1; ...; tn]] returns tn -> tn-1 -> ... -> t0
      for non-empty exp Lists (n >= 0) *)
  let makeArr (tl) =
    match tl with
	(t0 :: ts) -> List.fold_left (fun s -> fun t -> mkArr (t, s)) t0 ts in
    
(*
  (** [makeArr [t0; t1; ...; tn]] returns tn -> tn-1 -> ... -> t0
      for non-empty exp Lists (n >= 0) *)
  let rec makeArr (tl) =
    match tl with
	t :: s :: ts -> makeArr (mkArr (s, t) :: ts)
      | [t] -> t in
*)

  (* [parseRParen : (token -> 'a) -> token -> 'a] *)
  let parseRParen (p: token -> 'a) (c: token) =
    match c with
	Kwd ")" -> cont' p
      | _ -> raise (Fail ("closing parenthesis ')' expected")) in

  (* [parseColon : (token -> 'a) -> token -> 'a] *)
  let parseColon (p: token -> 'a) (c: token) =
    match c with
	Kwd ":" -> cont' p
      | _ -> raise (Fail ("colon ':' expected")) in

  (* [parseIdent: (string -> token -> 'a) -> token -> 'a] *)
  let parseIdent (p: string -> token -> 'a) (c: token) =
    match c with
	Ident x -> cont' (p x)
      | _ -> raise (Fail ("identifier expected"))
  in
(*
  let closeParen p = parseRParen (cont' p) in

  let removeRParen () =
    match Stream.peek (s) with
	Some (Kwd ")") -> Stream.junk (s)
      | _ -> raise (Fail ("closing parenthesis ')' expected")) in
*)

  (** Kinds *)

  (* [parseKind : (exp -> token -> 'a) -> token -> 'a] *)
  let rec parseKind (p: exp -> token -> 'a) = 
    parseKindAtom 
      (fun t -> parseMoreKinds ([t]) 
	 (fun tl -> p (makeArr tl)))
      
  (* [parseMoreKinds: exp list -> (exp list -> token -> 'a) -> token -> 'a] *)
  and parseMoreKinds (ts: exp list) (p: exp list -> token -> 'a) (c: token) = 
    match c with
	Kwd "->" ->
          cont' (parseKindAtom 
		   (fun t -> parseMoreKinds (t::ts) p))
      | _ -> p ts c
	  
  (* [parseKindAtom: (exp -> token -> 'a) -> token -> 'a] *)
  and parseKindAtom (p: exp -> token -> 'a) (c: token) = 
    match c with
	Kwd "(" -> cont' (parseKind 
			    (fun t -> parseRParen (p t)))
      | Kwd "*" -> cont' (p mkType)
      | _ -> raise (Fail ("kind expected")) in


  (** Types *)

  (** [mkVar: Clos.surr * string -> exp]
      Converts var identifier into de Bruijn index.
      Works for both local variables and global definitions. *)
  let mkVar (ids, a) = 
    try
      Var (Clos.index (ids, a)) 
    with
	UnknownIdent x -> raise (Fail ("unknown identifier '"^x^"'"))
  in

  (* [parseTy: (exp -> surr -> token -> 'a) -> surr -> token -> 'a] *)
  let rec parseTy (p: exp -> surr -> token -> 'a) = 
    parseNeuTy 
      (fun t -> parseMoreArrs [t] 
	 (fun tl  -> p (makeArr tl)))

  (* [parseMoreArrs: exp list -> (exp list -> surr -> token -> 'a) -> surr -> token -> 'a] *)
  and parseMoreArrs (ts: exp list) (p: exp list -> surr -> token -> 'a) (ids: surr) (c: token) = 
    match c with
      Kwd "->" ->  cont' (parseNeuTy 
			    (fun t -> parseMoreArrs (t::ts) p)
			    ids)
      | _ -> p ts ids c

  (* [parseNeuTy: (exp -> surr -> token -> 'a) -> surr -> token -> 'a] *)
  and parseNeuTy (p: exp -> surr -> token -> 'a) =  
    parseTyAtom' 
      (fun t -> parseMoreApps t p)

  (* [parseMoreApps: exp -> (exp -> surr -> token -> 'a) -> surr -> token -> 'a] *)
  and parseMoreApps (t0: exp) (p: exp -> surr -> token -> 'a) = 
    parseTyAtom 
      (fun t1 -> parseMoreApps (mkApp (t0, t1)) p) 
      (p t0)

  (* [parseTyAtom: (exp -> surr -> token -> 'a) -> surr -> token -> 'a] *)
  and parseTyAtom (p: exp -> surr -> token -> 'a) 
                  (fail: surr -> token -> 'a) (ids: surr) (c: token) = 
    match c with
	Ident a -> cont' (p (mkVar (ids, a)) ids)
      | Kwd "(" -> cont' (parseTy 
			    (fun t -> fun ids -> parseRParen (p t ids)) 
			    ids)
      | Kwd "." -> cont' (parseTy p ids)
      | Kwd "\\" -> cont' (parseIdent 
			     (fun a ->  
				parseTyAtom' 
				(fun t -> fun ids' -> p (mkLam (a, t)) ids)
				(Clos.pushLit' (a, ids))))  
      | Kwd "forall" -> cont' (parseIdent 
				 (fun a -> parseForall a p ids))
      | _ -> fail ids c

  and parseTyAtom' p = parseTyAtom p 
      (fun ids -> fun c -> raise (Fail ("type expected")))

  (* [parseForall: string -> (exp -> surr -> token -> 'a) -> surr -> token -> 'a] *)
  and parseForall (a: string) (p: exp -> surr -> token -> 'a) (ids: surr) (c:token) = 
    let ids' = Clos.pushLit' (a, ids) in
      match c with 
	  Kwd ":" -> cont' (parseKind 
			      (fun kappa -> parseTyAtom'
				 (fun t -> fun _ -> p (mkForall (a, kappa, t)) ids)
				 ids'))
	| _ -> parseTyAtom' (fun t -> fun _ -> p (mkForall (a, mkType, t)) ids) ids' c
  in

(* does not work: let parseType = parseTy (Top.tySurr()) *)
(*
let parseOnlyType = parser [< t = parseTy (Global.tySurr()) ; _ = Stream.empty >] -> t 
let stringToType s = parseOnlyType (lexer (Stream.of_string s)) 
*)

  (** Terms *)

  (* Grammar:
     
    Tm   ::= Atom Tms
    Tms  ::= Atom Tms
	   | <nothing>
    Atom ::= Id
	   | ( Tm )
	   | . Tm
	   | \ Id Atom
  *)
 
  (* [parseTm:  (exp -> surr -> token -> 'a) -> surr -> token -> 'a] *)
  let rec parseTm p = 
    parseTmAtom' (fun t -> parseMoreTms t p)

  (* [parseMoreTms : exp ->  (exp -> surr -> token -> 'a) -> surr -> token -> 'a] *)
  and parseMoreTms t0 p = 
    parseTmAtom 
      (fun t1 -> parseMoreTms (mkApp (t0, t1)) p)
      (p t0)

  (* [parseTmAtom: (exp -> surr -> token -> 'a) -> (surr -> token -> 'a) -> surr -> token -> 'a] *)
  and parseTmAtom p fail ids c = 
    match c with
	Ident a -> cont' (p (mkVar (ids, a)) ids)
      | Kwd "(" -> cont' (parseTm (fun t -> fun ids -> parseRParen (p t ids)) ids)
      | Kwd "." -> cont' (parseTm p ids)
      | Kwd "\\" -> cont' (parseIdent 
			     (fun a -> parseTmAtom' 
				(fun t -> fun ids' -> p (mkLam (a, t)) ids) 
				(Clos.pushLit' (a, ids)))) 
      | _ -> fail ids c

  and parseTmAtom' p = parseTmAtom p 
    (fun ids -> fun c -> raise (Fail ("term expected")))

  in

(* does not work: let parseTerm = parseTm (Top.tmSurr()) *)
(*
let parseOnlyTerm = parser 
    [< t = parseTm (Global.tmSurr()); _ = Stream.empty >] -> t 

let stringToTerm s = parseOnlyTerm (lexer (Stream.of_string s)) 
*)

  (** Declarations *)

  (* [parseTmDeclOrDef : string * exp -> (decl -> token -> 'a) -> token -> 'a] *)
  let parseTmDeclOrDef (x, tau) p (c) = 
    match c with
	Kwd "=" -> cont' (parseTm 
			    (fun t -> fun ids -> p (Decl.TmDef (x, tau, t))) 
			    (Global.tmSurr()))
      | _ -> p (Decl.TmDecl (x, tau)) c
  in

  (* [parseTyDeclOrDef : string * exp -> (decl -> token -> 'a) -> token -> 'a] *)
  let parseTyDeclOrDef (x, kappa) p (c) = 
    match c with
	Kwd "=" -> cont' 
	  (parseTy 
	     (fun tau -> fun ids -> p (Decl.TyDef (x, kappa, tau))) 
	     (Global.tySurr()))
      | _ -> p (Decl.TyDecl (x, kappa)) c

  in

  (* [parseDecl : (decl -> token -> 'a) -> token -> 'a] *)
  let parseDecl (p: Decl.decl -> token -> 'a) c =
    match c with
	Kwd "term" -> cont' 
	  (parseIdent 
	     (fun x -> parseColon
		(parseTy  
		   (fun tau -> fun ids -> parseTmDeclOrDef (x, tau) p)
		   (Global.tySurr()))))
      | Kwd "type" -> cont' 
	  (parseIdent 
	     (fun x -> parseColon 
		(parseKind (fun kappa -> parseTyDeclOrDef (x, kappa) p))))
      | _ -> 
	  parseTm 
	  (fun t -> fun ids -> parseColon
	     (parseTy (fun tau -> fun ids -> p (Decl.Chk (t, tau)))
		(Global.tySurr())))
	  (Global.tmSurr()) c
  in
  let parseSemi p c =
    match c with 
	Kwd ";" -> p ()
      | _ -> raise (Fail ("semicolon ';' expected"))
  in

  let rec parseLine () =
    cont 
      (parseDecl (fun d -> parseSemi (fun () -> Top.doDecl d; parseLine ())))
      (fun () -> ())

  in (* begin parseFile *)
    parseLine ()

(*
let stringToDecl  s = parseDecl (lexer (Stream.of_string s))
let stringToDecls s = parseLines (lexer (Stream.of_string s))
*)
