(* sig surr = cxt + env *)

(** Context and environment handling. *)

open Exp
   
type cxt 
type env 
type preExp  (* unlifted exp *)

(* CONTEXTS *)
      
(* Constructors *)

(* Contexts are dynamical, with discharging              *)
(* indices are relative and positive                     *)
(* when retriving a variable decl, type has to be lifted *)
  
val cxt_empty : cxt
val cxt_pushLit : string * preExp * cxt -> cxt
val cxt_pushLit' : string * cxt -> cxt

(* care for unique names *)
val nextInst : (env * cxt) * string -> string
val cxt_push      : string * preExp * (env * cxt) -> cxt
val cxt_pushTmVar : string * (cxt * exp) * (env * cxt) -> cxt
val cxt_pushTyVar : string * exp * (env * cxt) -> cxt
val cxt_push'     : string * (env * cxt) -> cxt

val pack  : cxt * exp -> preExp  

(** Example:

   Let   gR |- tau : type  
   Then 
         gL, x:tau 
   is implemented by
         push (x, pack (tau, gR), gL)
*)

(* Methods *)
  
val cxt_length  : cxt -> int

val cxt_lookup : cxt * int -> string * preExp
val unpack   : preExp * cxt -> exp

 (** [promote (gamma, t, gamma') = t']
    - [gamma]:  old context
    - [gamma']: new context
    - raises [Pie.Occurrence] if strengthening not possible
 *)
val promote : cxt * exp * cxt -> exp

(* unused
 (** [tmVarPromote (gamma, t, gamma') = t']  
    Promotion (lifting) of term variables in Church-term [t]
    - [gamma]:  old context
    - [gamma']: new context
    - raises [Error] if strengthening requested
 *)
val tmVarPromote : cxt * exp * cxt -> exp

 (** [tyVarPromote (gamma, t, gamma') = t']  
    Promotion (lifting) of type variables in Church-term [t]
    - [gamma]:  old context
    - [gamma']: new context
    - raises [Error] if strengthening requested
 *)
val tyVarPromote : cxt * exp * cxt -> exp
*)

val churchPromote : exp * (env * cxt) * cxt * (env * cxt) * cxt -> exp

(* unused
(** Church-style term abstraction *)

val tmAbstract : cxt * cxt * exp -> cxt * exp
(** [tmAbstract (gL, gR, e) = (gL', mkTmAbs (..., e))] *)
*)

 (** [lookupTmVar (gL, i, gR) = (x, tau)]  such that [gR |- tau : *] *)
val cxt_lookupTmVar : cxt * int * cxt -> string * exp

 (** [lookupTyVar (gL, i) = (a, kappa)] *)
val cxt_lookupTyVar : cxt * int -> string * exp
  
val cxt_index : cxt * string -> int option

(* only used internally:
val cxt_varsToString : cxt -> string
*)

(* ENVIRONMENTS *)
  
(* Unification variables *)
   
type cell =
    Name of string * cxt * exp             (* (X, Gamma, kappa) *)
  | Value of exp * cxt * exp               (* (tau, Gamma, kappa) *)
  | ChName of string * cxt * cxt * exp     (** Church cell (x, gL, gR, tau) *)
  | ChValue of exp * cxt * cxt * exp       (** (e, gL, gR, tau) *)

(* Environments *)

(* indices are negative and absolute, lifting never required *)
(* no discharging *)

type entry =
    Decl of string * exp                       (* c : tau               *)
  | Def of string * exp * exp * bool           (* c : tau = e (strict?) *)
  | Cell of cell ref

    
val env_empty : env

val env_pushDecl  : string * exp * env -> env
 (** [pushDecl (x, tau, theta)] *)

val env_pushDef   : string * exp * exp * bool * env -> env
 (** [pushDef (x, tau, t, strict, theta)] *)

val env_pushEVar  : string * cxt * exp * env -> env 
val env_introEVarName : string * cxt * exp * env * exp -> env * exp * exp * string
 (** [introEVar (x, gamma, kappa, theta, sigma) = (theta', sigma', ex, x')]
    - introduce EVar [x] in [sigma] as Var 0
    - push [(x, gamma, kappa)] onto [theta]
    - return Church-term [ex = x]
    - return name [x'] of new EVar
  *)  
val env_introEVarX: string * cxt * exp * env * exp -> env * exp * exp
 (** [introEVar (x, gamma, kappa, theta, sigma) = (theta', sigma', e)]
    - introduce EVar [x] in [sigma] as Var 0
    - push [(x, gamma, kappa)] onto [theta]
    - return Church-term [e = x]
  *)  
val env_introEVar : string * cxt * exp * env * exp * exp -> env * exp * exp
 (** [introEVar (x, gamma, kappa, theta, sigma, e) = (theta', sigma', e')]
    - introduce EVar [x] in [sigma] as Var 0
    - push [(x, gamma, kappa)] onto [theta]
    - return Church-term [e' = e x]
  *)  

val env_promote : env * env * cxt -> env
(** [env_promote (theta, thStop, gNew)] 

  Preliminaries: [theta = append (th, thStop)]

  Promotes all uninstantiated EVars in [th] to new context [gNew].
  Warning! Has side effects!
*)

val env_newChVar : string * cxt * cxt * exp * env -> env * exp * cell ref
  (** [newEVar (x, gL, gR, sigma, theta) = (theta', e, r)]
    - push [(x, gL, gR, sigma)] onto [theta]
    - return [e = x] and [r] as reference to the new Church-term cell
   *)

val env_newTmVar' : string * cxt * exp * env -> env * exp * cell ref
  (** [newEVar (x, gamma, sigma, theta) = (theta', e, r)]
    - push [(x, gamma, sigma)] onto [theta]
    - return [e = x] and [r] as reference to the new cell
   *)

val env_newEVar' : string * cxt * exp * env -> env * exp * cell ref
  (** [newEVar (x, gamma, kappa, theta) = (theta', sigma, r)]
    - push [(x, gamma, kappa)] onto [theta]
    - return [sigma = x] and [r] as reference to the new cell
   *)

(* no longer needed: *)
val env_introEVar' : string * cxt * exp * env * exp * (exp list -> exp) 
  -> env * exp * (exp list -> exp)
 (** [introEVar' (x, gamma, kappa, theta, sigma, f) = (theta', sigma', f')]
    - introduce EVar [x] in [sigma] as Var 0
    - push [(x, gamma, kappa)] onto [theta]
    - return delayed Church-term [f' = fun es -> (f es) x]
  *)  
val env_newEVar  : string * cxt * exp * env -> env * exp 
  (** [newEVar (x, gamma, kappa, theta) = (theta', sigma)]
    - push [(x, gamma, kappa)] onto [theta]
    - return [sigma = x]
   *)

val env_lookup : env * int -> entry
val env_index  : env * string -> int option
